<?php

class Pin extends ActiveRecord\Model implements CodigoAcessoInterface {

	private $nameFieldCodigoAcesso = 'pin';
	private $maxNumberTimesRecovered = 1;
	private $tamanhoCodigoAcesso = 6;
	private $metodoRecuperacaoCodigoAcesso = ['recuperarSimples', 'recuperarByMaxNumberTimesByDay'];
	private $metodoAutenticacao = ['autenticarSimples', 'autenticarComLimiteNroAutenticacoes', 'autenticarComLimiteTempo',];
	private $tempoAcessoLiberado = '120'; // por padrao, 120 minutos (2h)

	static $validates_presence_of = [
		['user_id', 'message' => 'deve ser fornecido!'],
		['mac', 'message' => 'deve ser fornecido!'],
		['pin', 'message' => 'deve ser fornecido!']
	];

	public function setMaxNumberTimesRecovered($maxNumberTimesRecovered) {
		$this->maxNumberTimesRecovered = $maxNumberTimesRecovered;
	}

	public function setTamanhoCodigoAcesso($tamanhoCodigoAcesso) {
		$this->tamanhoCodigoAcesso = $tamanhoCodigoAcesso;
	}

	public function getTempoAcessoLiberado() {
		return $this->tempoAcessoLiberado;
	}

	public function getMaxNumberTimesRecovered() {
		return $this->maxNumberTimesRecovered;
	}

  	public function temCodigoAcesso(array $keys) {
  		$conditions = static::createStringConditions($keys);
  		return static::exists(['conditions' => $conditions]);
  	}

	public function gerarCodigoAcesso() {
		$allCodigosAcessos = $this->getAllCodigosAcessos();

		do {
			$newCodigoAcesso = substr(uniqid(), $this->tamanhoCodigoAcesso * -1);
		} while(in_array($newCodigoAcesso, $allCodigosAcessos));

		return strtoupper($newCodigoAcesso);
	}

	private function getAllCodigosAcessos() {
		$resultado = [];
		foreach(static::all() as $codigoAcesso) {
			$resultado[] = $codigoAcesso->{$this->nameFieldCodigoAcesso};
		}

		return $resultado;
	}

	private function recuperarSimples(array $keys) {
		$conditions = static::createStringConditions($keys);
		$resultado = static::find('first', ['conditions' => $conditions]);

		if($resultado) {
			$this->atualizarRecuperacaoCodigoAcesso($resultado);
			return $resultado->{$this->nameFieldCodigoAcesso};
		}

		return false;
	}

	private function atualizarRecuperacaoCodigoAcesso(Pin $obj) {

		if($obj->{$this->nameFieldCodigoAcesso} == '') { // assegura que o objeto nao eh um novo objeto, mas sim um objeto recuperado
			return false;
		}

		$obj->total_number_times_recovered += 1;
		$obj->last_recovered_at = strftime("%Y-%m-%d %T");
		$obj->save(false);
	}

	private function recuperarByMaxNumberTimesByDay($keys) {
		$maxNumberTimes = $this->getMaxNumberTimesRecovered();
		$dataAtual = strftime("%Y-%m-%d");

		$conditions = static::createStringConditions($keys);
		$resultado = static::find('first', ['conditions' => $conditions]);

		$ok = false;

		if(!$resultado) {
			return false;
		}

		$dataLastRecovered = $resultado->last_recovered_at->format('Y-m-d');

		if($dataAtual != $dataLastRecovered) {
			$resultado->number_times_recovered = 1; // reseta contador de quantidade de vezes recuperado
			$ok = true;
		} else { // mesma data, entao confere o limite de vezes recuperado
			if($resultado->number_times_recovered < $maxNumberTimes) {
				$resultado->number_times_recovered += 1; // incrementa contador de quantidade de vezes recuperado
				$ok = true;
			}
		}

		if($ok) {
			$this->atualizarRecuperacaoCodigoAcesso($resultado);
			return $resultado->{$this->nameFieldCodigoAcesso};
		}

		return false;
	}

	public function recuperarCodigoAcesso(array $keys, $tipoMetodoRecuperacaoCodigoAcesso = 0) {
		$metodoRecuperacaoCodigoAcesso = $this->getMetodoRecuperacaoCodigoAcesso($tipoMetodoRecuperacaoCodigoAcesso);

		return call_user_func(array($this, $metodoRecuperacaoCodigoAcesso), $keys);
	}

	public function getMetodoRecuperacaoCodigoAcesso($tipoMetodoRecuperacaoCodigoAcesso) {
		return $this->metodoRecuperacaoCodigoAcesso[$tipoMetodoRecuperacaoCodigoAcesso];
	}

	private function createStringConditions(array $conditions) {
		$sqlCondition = '1';
		foreach($conditions as $key => $value) {
			$sqlCondition.=sprintf(" AND %s = '%s'", $key, $value);
		}

		return $sqlCondition;
	}

	private function getMetodoAutenticacao($tipoMetodoAutenticacao) {
		return $this->metodoAutenticacao[$tipoMetodoAutenticacao];
	}

	public function autenticar(array $dadosAutenticacao, $tipoMetodoAutenticacao = 0) {
		$metodoAutenticacao = $this->getMetodoAutenticacao($tipoMetodoAutenticacao);
		return call_user_func(array($this, $metodoAutenticacao), $dadosAutenticacao);
	}

	private function atualizarAuth(Pin $obj, array $dadosAdicionais = array()) {

		if($obj->{$this->nameFieldCodigoAcesso} == '') { // assegura que o objeto nao eh um novo objeto, mas sim um objeto recuperado
			return false;
		}

		// parte comum
		$obj->last_auth = strftime('%Y-%m-%d %T');
		$obj->total_number_auth += 1;

		foreach($dadosAdicionais as $key => $value) {
			$obj->$key = $value;
		}

		return $obj->save(false);
	}

	private function isFirstAuth(Pin $obj) {
		if($obj->{$this->nameFieldCodigoAcesso} == '') { // assegura que o objeto nao eh um novo objeto, mas sim um objeto recuperado
			return false;
		}

		return is_null($obj->last_auth); // PHP-AR instancia como NULL ao inves de DateTime quando o field esta marcado como '0000-00-00'
	}

	private function resetarAuth(Pin $obj) {
		if($obj->{$this->nameFieldCodigoAcesso} == '') { // assegura que o objeto nao eh um novo objeto, mas sim um objeto recuperado
			return false;
		}

		$currentDate = new DateTime();
		$interval = 'P1'.$obj->periodicity;

		$obj->last_auth->add(new DateInterval($interval));

		if($currentDate >= $obj->last_auth) {
			$obj->current_number_auth = 0; // reseta contador da autenticacao atual
			return $obj->save(false);
		}

		return false;
	}

	private function verificarLimiteAuth(Pin $obj) {
		if($obj->{$this->nameFieldCodigoAcesso} == '') { // assegura que o objeto nao eh um novo objeto, mas sim um objeto recuperado
			return false;
		}

		return $obj->current_number_auth < $obj->max_number_auth;
	}

	private function autenticarSimples(array $dadosAutenticacao) {
		$conditions = static::createStringConditions($dadosAutenticacao);
		$pin = static::find('first', ['conditions' => $conditions]);
        $this->tempoAcessoLiberado = $pin->time_to_live;

		if(!$pin) {
			return false;
		}

		return $this->atualizarAuth($pin);
	}

	private function autenticarComLimiteNroAutenticacoes(array $dadosAutenticacao) {
		$conditions = static::createStringConditions($dadosAutenticacao);
		$pin = static::find('first', ['conditions' => $conditions]);

		if(!$pin) {
			return false;
		}

		if($this->isFirstAuth($pin)) { // se for primeira autenticacao, entao libera acesso e atualiza os dados
			return $this->atualizarAuth($pin, ['current_number_auth' => $pin->current_number_auth + 1]);
		}

		// nao se trata de primeira autenticacao, entao verifica se pode resetar os contadores e posteriormente verifica os limites
		$this->resetarAuth($pin);
		if($this->verificarLimiteAuth($pin)) {
			return $this->atualizarAuth($pin, ['current_number_auth' => $pin->current_number_auth + 1]);
		}

		return false;
	}

    private function autenticarComLimiteTempo(array $dadosAutenticacao) {
        $conditions = static::createStringConditions($dadosAutenticacao);
        $pin = static::find('first', ['conditions' => $conditions]);

        if(!$pin) {
            return false;
        }

        if ($this->isFirstAuth($pin) || $this->renovouPeriodo($pin)) {
            $this->tempoAcessoLiberado = $pin->time_to_live;
            return $this->atualizarAuth($pin, ['expiry_time' => $this->calculateExpiryTime($pin), 'current_number_auth' => $pin->current_number_auth + 1]);
        }

        // nao renovou o periodo, entao verifica se ainda esta dentro do prazo de autenticacao, ou seja, verifica se o Pin ainda nao expirou
        if (! $this->isExpired($pin)) {
            $this->tempoAcessoLiberado = $this->calculateRemainingTime($pin);
            return $this->atualizarAuth($pin, ['current_number_auth' => $pin->current_number_auth + 1]);
        }

        return false; // nega autenticacao
    }

    private function renovouPeriodo(Pin $pin) {
        if (! $this->pinExistente($pin)) {
            return false;
        }

        $current = new DateTime(strftime('%Y-%m-%d'));
        $last = new DateTime($pin->last_auth->add(new DateInterval('P1'.$pin->periodicity))->format('Y-m-d'));

        return $current >= $last;
    }

    private function isExpired(Pin $pin) {
        if (! $this->pinExistente($pin)) {
            return true;
        }

        $current = new DateTime();
        return $current >= $pin->expiry_time;
    }

    private function calculateRemainingTime(Pin $pin) {
        if (! $this->pinExistente($pin) || $this->isExpired($pin)) {
            return (int) 0;
        }

        $current = new DateTime();
        $since_start = $pin->expiry_time->diff($current);

        $minutes = $since_start->days * 24 * 60;
        $minutes += $since_start->h * 60;
        $minutes += $since_start->i;
        return $minutes;
    }

    private function calculateExpiryTime(Pin $pin) {
        $current = new DateTime();

        if (! $this->pinExistente($pin)) {
            return $current->format('Y-m-d H:i:s');
        }

        $current->add(new DateInterval(sprintf('PT%sM', $pin->time_to_live)));
        return $current->format('Y-m-d H:i:s');
    }

    private function pinExistente(Pin $pin) {
        if ($pin->{$this->nameFieldCodigoAcesso} == '') {
            return false;
        }

        return true;
    }
}
